import Data.Map (Map)
import qualified Data.Map as Map

-- An evaluator that handles failure through a monad

-- pure lambda calculus
data Expr = 
    Var String       -- x
  | Lam String Expr  -- \ x. e
  | App Expr Expr    -- e1 e2
  | Number Integer   -- n
  | Add Expr Expr    -- e1 + e2
  deriving (Show)

data Val =
    Clo String Expr Env  -- \ x . e in env
  | NumberV Integer      -- n
  deriving (Show)

-- this is a type alias
type Env = Map String Val

coerceClo :: Val -> Maybe (String, Expr, Env)
coerceClo (Clo x e env) = Just (x, e, env)
coerceClo _ = Nothing

coerceNum :: Val -> Maybe Integer
coerceNum (NumberV n) = Just n
coerceNum _ = Nothing

-- Map.lookup :: (Eq k) => k -> Map k v -> Maybe v
-- Map.insert :: (Eq k) => k -> v -> Map k v -> Map k v
-- (=<<) :: (a -> Maybe b) -> Maybe a -> Maybe b
-- (=<<) is just a flipped version of (>>=), or bind
-- f =<< aM = do { a <- aM ; f a }
-- f =<< aM is morally "apply f to the result of aM"
-- a big step interpreter
eval :: Expr -> Env -> Maybe Val
eval (Var x) env = do
  Map.lookup x env
eval (Lam x body) env = do
  return (Clo x body env)
eval (App e1 e2) env = do
  (x, body, env') <- coerceClo =<< eval e1 env
  v2 <- eval e2 env
  eval body (Map.insert x v2 env')
eval (Number n) env = do
  return (NumberV n)
eval (Add e1 e2) env = do
  n1 <- coerceNum =<< eval e1 env
  n2 <- coerceNum =<< eval e2 env
  return (NumberV (n1 + n2))

-- if you 'runhaskell' this file it will print 'oh hai' signaling that this
-- file type-checks
main :: IO ()
main = putStrLn "oh hai"
