import Data.Map (Map)
import qualified Data.Map as Map

-- An evaluator that handles failure through a monad and state explicitly

-- pure lambda calculus
data Expr = 
    Var String       -- x
  | Lam String Expr  -- \ x. e
  | App Expr Expr    -- e1 e2
  | Number Integer   -- n
  | Add Expr Expr    -- e1 + e2
  | Ref Expr         -- ref e
  | Lookup Expr      -- !e
  | Assign Expr Expr -- e1 := e2
  deriving (Show)

data Val =
    Clo String Expr Env  -- \ x . e in env
  | NumberV Integer      -- n
  | Location Integer     -- l
  deriving (Show)

-- these are type aliases
type Env = Map String Val
type Heap = (Map Integer Val, Integer)

coerceClo :: Val -> Maybe (String, Expr, Env)
coerceClo (Clo x e env) = Just (x, e, env)
coerceClo _ = Nothing

coerceNum :: Val -> Maybe Integer
coerceNum (NumberV n) = Just n
coerceNum _ = Nothing

coerceLocation :: Val -> Maybe Integer
coerceLocation (Location i) = Just i
coerceLocation _ = Nothing

-- Map.lookup :: (Eq k) => k -> Map k v -> Maybe v
-- Map.insert :: (Eq k) => k -> v -> Map k v -> Map k v
-- (=<<) :: (a -> Maybe b) -> Maybe a -> Maybe b
-- (=<<) is just a flipped version of (>>=), or bind
-- f =<< aM = do { a <- aM ; f a }
-- f =<< aM is morally "apply f to the result of aM"
-- a big step interpreter
eval :: Expr -> Env -> Heap -> Maybe (Val, Heap)
eval (Var x) env heap = do
  v <- Map.lookup x env
  return (v,heap)
eval (Lam x body) env heap = do
  return (Clo x body env, heap)
eval (App e1 e2) env heap = do
  (v1, heap') <- eval e1 env heap
  (v2, heap'') <- eval e2 env heap'
  (x, body, env') <- coerceClo v1
  eval body (Map.insert x v2 env') heap''
eval (Number n) env heap = do
  return (NumberV n, heap)
eval (Add e1 e2) env heap = do
  (v1, heap') <- eval e1 env heap
  (v2, heap'') <- eval e2 env heap'
  n1 <- coerceNum v1
  n2 <- coerceNum v2
  return (NumberV (n1 + n2), heap'')
eval (Ref e) env heap = do
  (v, (heapMap, heapNext)) <- eval e env heap
  let heap' = (Map.insert heapNext v heapMap, heapNext + 1)
  return (Location heapNext, heap')
eval (Lookup e) env heap = do
  (v, heap') <- eval e env heap
  i <- coerceLocation v
  v' <- Map.lookup i (fst heap')
  return (v',heap')
eval (Assign e1 e2) env heap = do
  (v1, heap') <- eval e1 env heap
  i <- coerceLocation v1
  (v2, (heapMap, heapNext)) <- eval e2 env heap'
  let heapMap' = Map.insert i v2 heapMap
  return (v2, (heapMap',heapNext))

-- if you 'runhaskell' this file it will print 'oh hai' signaling that this
-- file type-checks
main :: IO ()
main = putStrLn "oh hai"
