import Data.Map (Map)
import qualified Data.Map as Map

-- An evaluator that handles failure and state through a monad

-- pure lambda calculus
data Expr = 
    Var String       -- x
  | Lam String Expr  -- \ x. e
  | App Expr Expr    -- e1 e2
  | Number Integer   -- n
  | Add Expr Expr    -- e1 + e2
  | Ref Expr         -- ref e
  | Lookup Expr      -- !e
  | Assign Expr Expr -- e1 := e2
  deriving (Show)

data Val =
    Clo String Expr Env  -- \ x . e in env
  | NumberV Integer      -- n
  | Location Integer     -- l
  deriving (Show)

-- these are type aliases
type Env = Map String Val
type Heap = (Map Integer Val, Integer)

----------

-- the MaybeState Monad
-- a computation with two effects:
-- * it could fail
-- * it modifies a single cell of state of type 'Heap'
data MaybeState a = MaybeState (Heap -> Maybe (a, Heap))

-- execute the effect
runMaybeState :: MaybeState a -> Heap -> Maybe (a, Heap)
runMaybeState (MaybeState f) heap = f heap

-- making MaybeState a monad
instance Monad MaybeState where
  return a = MaybeState (\ heap -> Just (a, heap))
  -- (>>=) :: MaybeState a -> (a -> MaybeState b) -> MaybeState b
  -- there is a more compact way to write this, but it's drawn out to show the
  -- types
  (>>=) (MaybeState aM) k = MaybeState heapK
    -- aM :: Heap -> Maybe (a, Heap)
    where
      -- heapK :: Heap -> Maybe (a, Heap)
      heapK heap =
        case aM heap of
          Nothing -> Nothing
          Just (a,heap') -> 
            let (MaybeState bM) = k a 
            -- bM :: Heap -> Maybe (b, Heap)
            in bM heap'

-- monadic effects that MaybeState provides

failure :: MaybeState a
failure = MaybeState (\ heap -> Nothing)

get :: MaybeState Heap
get = MaybeState (\ heap -> Just (heap, heap))

put :: Heap -> MaybeState ()
put heap = MaybeState (\ oldHeap -> Just ((), heap))

liftMaybe :: Maybe a -> MaybeState a
liftMaybe Nothing = MaybeState (\ heap -> Nothing)
liftMaybe (Just a) = MaybeState (\ heap -> Just (a, heap))

----------

-- Helpers, all built with failure, get and put (they don't peek into how
-- MaybeState works)

coerceClo :: Val -> MaybeState (String, Expr, Env)
coerceClo (Clo x e env) = return (x, e, env)
coerceClo _ = failure

coerceNum :: Val -> MaybeState Integer
coerceNum (NumberV n) = return n
coerceNum _ = failure

coerceLocation :: Val -> MaybeState Integer
coerceLocation (Location i) = return i
coerceLocation _ = failure

malloc :: Val -> MaybeState Integer
malloc v = do
  (heapMap, heapNext) <- get
  put (Map.insert heapNext v heapMap, heapNext + 1)
  return heapNext

heapLookup :: Integer -> MaybeState Val
heapLookup l = do
  (heapMap, _) <- get
  liftMaybe (Map.lookup l heapMap)

heapUpdate :: Integer -> Val -> MaybeState ()
heapUpdate l v = do
  (heapMap, heapNext) <- get
  put (Map.insert l v heapMap, heapNext)

-- Map.lookup :: (Eq k) => k -> Map k v -> Maybe v
-- Map.insert :: (Eq k) => k -> v -> Map k v -> Map k v
-- (=<<) :: (a -> Maybe b) -> Maybe a -> Maybe b
-- (=<<) is just a flipped version of (>>=), or bind
-- f =<< aM = do { a <- aM ; f a }
-- f =<< aM is morally "apply f to the result of aM"
-- a big step interpreter
-- returns a monadic computation
evalM :: Expr -> Env -> MaybeState Val
evalM (Var x) env = do
  liftMaybe (Map.lookup x env)
evalM (Lam x body) env = 
  return (Clo x body env)
evalM (App e1 e2) env = do
  (x, body, env') <- coerceClo =<< evalM e1 env
  v2 <- evalM e2 env
  evalM body (Map.insert x v2 env')
evalM (Number n) env = do
  return (NumberV n)
evalM (Add e1 e2) env = do
  n1 <- coerceNum =<< evalM e1 env
  n2 <- coerceNum =<< evalM e2 env
  return (NumberV (n1 + n2))
evalM (Ref e) env = do
  v <- evalM e env
  l <- malloc v
  return (Location l)
evalM (Lookup e) env = do
  heapLookup =<< coerceLocation =<< evalM e env
evalM (Assign e1 e2) env = do
  l <- coerceLocation =<< evalM e1 env
  v <- evalM e2 env
  heapUpdate l v
  return v

-- to fully evaluate, build the monadic computation and 'run' it
eval :: Expr -> Env -> Heap -> Maybe (Val, Heap)
eval e env heap = runMaybeState (evalM e env) heap

-- if you 'runhaskell' this file it will print 'oh hai' signaling that this
-- file type-checks
main :: IO ()
main = putStrLn "oh hai"
